/*
 * Copyright (C) 2017, 2019 Иван Вахромеев <ivanalert@mail.ru>
 *
 * This file is part of VkUnmask.

 * VkUnmask is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VkUnmask is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VkUnmask. If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <string>
#include <vector>
#include <algorithm>
#include <getopt.h>

using namespace std;

static const string all = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMN0PQRSTUVWXYZO123456789+/=";

vector<string> splitString(const string &str, const string &delimeter)
{
    vector<string> result;
    if (!str.empty() && !delimeter.empty())
    {
        string::size_type begin = 0, match;
        while ((match = str.find(delimeter, begin)) != string::npos)
        {
            string::size_type subSize = match - begin;
            if (subSize)
                result.push_back(str.substr(begin, subSize));

            begin = match + delimeter.size();
        }
        if (begin < str.size())
            result.push_back(str.substr(begin));
    }

    return result;
}

string aFunc(const string &str)
{
    if (str.empty() || str.size() % 4 == 1)
        return string();

    string result;
    string::size_type i, e, o;
    i = e = o = 0;

    for (string::size_type s = 0, size = str.size(); s < size; ++s)
    {
        e = all.find(str[s]);
        if (e != string::npos)
        {
            if (o % 4)
                i = 64 * i + e;
            else
                i = e;

            if (o++ % 4)
                result += char(255 & i >> (-2 * o & 6));
        }
    }

    return result;
}

vector<string::size_type> sFunc(const string &str, string::size_type number)
{
    string::size_type size = str.size();
    vector<string::size_type> result(size, 0);
    if (!result.empty())
    {
        string::size_type i = size;
        while (i--)
        {
            number = (size * (i + 1) ^ number + i) % size;
            result[i] = number;
        }
    }

    return result;
}

void rDecode(string &str, string::size_type i)
{
    string o = all + all;
    string::size_type e, s;
    s = str.size() - 1;
    while (s != string::npos)
    {
        e = o.find(str[s]);
        if (e != string::npos)
        {
            if (e >= i)
                str[s] = o[e - i];
            else
                str[s] = o[o.size() - (i - e)];
        }

        --s;
    }
}

void sDecode(string &str, string::size_type number)
{
    if (!str.empty())
    {
        vector<string::size_type> o = sFunc(str, number);
        for (string::size_type i = 1, size = str.size(); i < size; ++i)
        {
            string::size_type index = o.at(size - 1 - i);
            if (index >= size)
                index = size - 1;

            std::swap(str[i], str[index]);
        }
    }
}

string xDecode(const string &str1, const string &str2)
{
    string result;
    char ch = str2[0];
    for (string::size_type i = 0; i < str1.size(); ++i)
    {
        char currCh(str1[i]);
        result += currCh ^ ch;
    }

    return result;
}

int main(int argc, char *argv[])
{
    //Get options i - id, l - link.
    string idStr, link;
    static const struct option opts[] =
    {
        { "id", required_argument, nullptr, 'i' },
        { "link", required_argument, nullptr, 'l'}
    };
    int opt = 0, optIndex = 0;
    while ((opt = getopt_long(argc, argv, "i:l:", opts, &optIndex)) != -1)
    {
        switch (opt)
        {
        case 'i':
            idStr.assign(optarg);
            break;
        case 'l':
            link.assign(optarg);
            break;
        }
    }

    //Check options.
    unsigned long id = 0;
    try
    {
        id = stoul(idStr);
    }
    catch (exception_ptr ex)
    {
        cout << "Invalid id value!" << endl;
        return 1;
    }

    if (link.empty())
    {
        cout << "Empty link value!" << endl;
        return 1;
    }

    vector<string> mask = splitString(link, "?extra=");
    if (mask.size() > 1)
    {
        mask = splitString(mask[1], "#");
        if (mask.size() > 1)
        {
            string leftMask = aFunc(mask[0]);
            mask = splitString(aFunc(mask[1]), string(1, char(9)));

            vector<string> decodeStrs;
            vector<string>::const_reverse_iterator it = mask.crbegin();
            vector<string>::const_reverse_iterator end = mask.crend();
            while (it != end)
            {
                decodeStrs = splitString(*it, string(1, char(11)));
                if (decodeStrs.size() > 1)
                {
                    switch (decodeStrs[0][0])
                    {
                    case 'i':
                        sDecode(leftMask, stoul(decodeStrs[1]) ^ id);
                        break;
                    case 'r':
                        reverse(leftMask.begin(), leftMask.end());
                        break;
                    case 's':
                        sDecode(leftMask, stoul(decodeStrs[1]));
                        break;
                    case 'x':
                        leftMask = xDecode(leftMask, decodeStrs[1]);
                        break;
                    }

                    ++it;
                }
            }

            cout << leftMask << endl;
            return 0;
        }
    }

    cout << "Can't decode the link!" << endl;
    return 0;
}
